@@ -39,7 +39,7 @@ del os, sys, path
 # ... when internals _really_ fail & we want to know why
 def _dbg(text):
     import sys
-    from bam_utils.system import colorize
+    from bam.utils.system import colorize
     if type(text) is bytes:
         text = text.decode('utf-8')
     sys.__stdout__.write(colorize(text, color='red') + "\n")
@@ -47,11 +47,11 @@ def _dbg(text):
 
 
 def _relpath_remap(
-        path_src,
-        base_dir_src,
-        fp_basedir,
-        blendfile_src_dir_fakeroot,
-        ):
+        path_src: str,
+        base_dir_src: str,
+        fp_basedir: str,
+        blendfile_src_dir_fakeroot: str=None,
+        ) -> (str, str):
 
     import os
 
@@ -65,7 +65,18 @@ def _relpath_remap(
                 raise Exception("Internal error 'path_src' -> %r must be absolute" % path_src)
 
     path_src = os.path.normpath(path_src)
-    path_dst = os.path.relpath(path_src, base_dir_src)
+    if os.name != "nt":
+        path_dst = os.path.relpath(path_src, base_dir_src)
+    else:
+        # exception for windows, we need to support mapping between drives
+        try:
+            path_dst = os.path.relpath(path_src, base_dir_src)
+        except ValueError:
+            # include the absolute path when the file is on a different drive.
+            path_dst = os.path.relpath(
+                    os.path.join(base_dir_src, b'__' + path_src.replace(b':', b'\\')),
+                    base_dir_src,
+                    )
 
     if blendfile_src_dir_fakeroot is None:
         # /foo/../bar.png --> /foo/__/bar.png
@@ -99,7 +110,19 @@ def pack(
         #    os.path.dirname(blendfile_src)
         # but in some cases we wan't to use a path higher up.
         # base_dir_src,
-        blendfile_src, blendfile_dst, mode='FILE',
+        blendfile_src, blendfile_dst,
+
+        # the path to the top directory of the project's repository.
+        # the packed archive will reproduce the exact same hierarchy below that base path
+        # if set to None, it defaults to the given blendfile_src's directory.
+        # especially useful when used together with the warn_remap_externals option.
+        repository_base_path=None,
+
+        # type of archive to produce (either ZIP or plain usual directory).
+        mode='ZIP',
+
+        # optionally pass in the temp dir
+        base_dir_dst_temp=None,
         paths_remap_relbase=None,
         deps_remap=None, paths_remap=None, paths_uuid=None,
         # load every libs dep, not just used deps.
@@ -121,15 +144,21 @@ def pack(
         blendfile_src_dir_fakeroot=None,
 
         # Read variations from json files.
-        use_variations=True,
+        use_variations=False,
 
         # do _everything_ except to write the paths.
         # useful if we want to calculate deps to remap but postpone applying them.
         readonly=False,
+        # Warn when we found a dependency external to given repository_base_path.
+        warn_remap_externals=False,
         # dict of binary_edits:
         # {file: [(ofs, bytes), ...], ...}
         # ... where the file is the relative 'packed' location.
         binary_edits=None,
+
+        # Filename filter, allow to exclude files from the pack,
+        # function takes a string returns True if the files should be included.
+        filename_filter=None,
         ):
     """
     :param deps_remap: Store path deps_remap info as follows.
@@ -149,8 +178,12 @@ def pack(
     #   also prevents cyclic loops from crashing.
 
     import os
+    import sys
 
-    from bam.utils.system import colorize
+    if sys.stdout.isatty():
+        from bam.utils.system import colorize
+    else:
+        from bam.utils.system import colorize_dummy as colorize
 
     # in case this is directly from the command line or user-input
     blendfile_src = os.path.normpath(os.path.abspath(blendfile_src))
@@ -170,7 +203,8 @@ def pack(
     TEMP_SUFFIX = b'@'
 
     if report is None:
-        report = lambda msg: msg
+        def report(msg):
+            return msg
 
     yield report("%s: %r...\n" % (colorize("\nscanning deps", color='bright_green'), blendfile_src))
 
@@ -178,15 +212,25 @@ def pack(
         import time
         t = time.time()
 
-    base_dir_src = os.path.dirname(blendfile_src)
+    base_dir_src = os.path.dirname(blendfile_src) if repository_base_path is None \
+                                                  else os.path.normpath(os.path.abspath(repository_base_path))
     base_dir_dst = os.path.dirname(blendfile_dst)
     # _dbg(blendfile_src)
     # _dbg(blendfile_dst)
 
-    if mode == 'ZIP':
-        base_dir_dst_temp = os.path.join(base_dir_dst, b'__blendfile_temp__')
-    else:
-        base_dir_dst_temp = os.path.join(base_dir_dst, b'__blendfile_pack__')
+    if base_dir_dst_temp is None:
+        # Always try to pack using a unique folder name.
+        import uuid
+
+        suf = 'temp' if mode == 'ZIP' else 'pack'
+
+        while True:
+            unique = uuid.uuid4().hex
+            name = '__blendfile_%s_%s__' % (unique, suf)
+            base_dir_dst_temp = os.path.join(base_dir_dst, name.encode('ascii'))
+
+            if not os.path.exists(base_dir_dst_temp):
+                break
 
     def temp_remap_cb(filepath, rootdir):
         """
@@ -306,7 +350,7 @@ def pack(
         fp_blend = os.path.join(fp.basedir, fp_blend_basename)
 
         if fp_blend_basename_last != fp_blend_basename:
-            yield report("  %s:       %s\n" % (colorize("blend", color='blue'), fp_blend))
+            yield report("  %s:       %r\n" % (colorize("blend", color='blue'), fp_blend))
             fp_blend_basename_last = fp_blend_basename
 
             if binary_edits is not None:
@@ -327,6 +371,13 @@ def pack(
         path_src = blendfile_path_walker.utils.abspath(path_rel, fp.basedir)
         path_src = os.path.normpath(path_src)
 
+        if warn_remap_externals and b".." in os.path.relpath(path_src, base_dir_src):
+            yield report("  %s:   %r\n" % (colorize("non-local", color='bright_yellow'), path_src))
+
+        if filename_filter and not filename_filter(path_src):
+            yield report("  %s:     %r\n" % (colorize("exclude", color='yellow'), path_src))
+            continue
+
         # apply variation (if available)
         if use_variations:
             if blendfile_levels_dict_curr:
@@ -372,8 +423,8 @@ def pack(
                 _dst_dir = os.path.dirname(path_dst)
                 path_copy_files.update(
                         {(os.path.join(_src_dir, f), os.path.join(_dst_dir, f))
-                        for f in file_list
-                        })
+                         for f in file_list
+                         })
                 del _src_dir, _dst_dir
 
         if deps_remap is not None:
@@ -405,9 +456,11 @@ def pack(
     if paths_remap is not None:
 
         if paths_remap_relbase is not None:
-            relbase = lambda fn: os.path.relpath(fn, paths_remap_relbase)
+            def relbase(fn):
+                return os.path.relpath(fn, paths_remap_relbase)
         else:
-            relbase = lambda fn: fn
+            def relbase(fn):
+                return fn
 
         for src, dst in path_copy_files:
             # TODO. relative to project-basepath
@@ -457,9 +510,13 @@ def pack(
         shutil.move(blendfile_dst_tmp, blendfile_dst)
         path_temp_files.remove(blendfile_dst_tmp)
 
-        # strip TEMP_SUFFIX
+        # strip TEMP_SUFFIX and move to the destination directory.
         for fn in path_temp_files:
-            shutil.move(fn, fn[:-len(TEMP_SUFFIX)])
+            dst_rel, _ = _relpath_remap(fn[:-len(TEMP_SUFFIX)], base_dir_dst_temp, base_dir_dst, None)
+            dst = os.path.join(base_dir_dst, dst_rel)
+            yield report("  %s: %r -> %r\n" % (colorize("moving", color='blue'), fn, dst))
+            os.makedirs(os.path.dirname(dst), exist_ok=True)
+            shutil.move(fn, dst)
 
         for src, dst in path_copy_files:
             assert(b'.blend' not in dst)
@@ -469,8 +526,11 @@ def pack(
                 yield report("  %s: %r\n" % (colorize("source missing", color='red'), src))
             else:
                 yield report("  %s: %r -> %r\n" % (colorize("copying", color='blue'), src, dst))
+                os.makedirs(os.path.dirname(dst), exist_ok=True)
                 shutil.copy(src, dst)
 
+        shutil.rmtree(base_dir_dst_temp)
+
         yield report("  %s: %r\n" % (colorize("written", color='green'), blendfile_dst))
 
     elif mode == 'ZIP':
@@ -484,7 +544,8 @@ def pack(
         zlib.Z_DEFAULT_COMPRESSION = compress_level
         _compress_mode = zipfile.ZIP_STORED if (compress_level == 0) else zipfile.ZIP_DEFLATED
         if _compress_mode == zipfile.ZIP_STORED:
-            is_compressed_filetype = lambda fn: False
+            def is_compressed_filetype(fn):
+                return False
         else:
             from bam.utils.system import is_compressed_filetype
 
@@ -545,13 +606,17 @@ def create_argparse():
             )
     parser.add_argument(
             "-m", "--mode", dest="mode", metavar='MODE', required=False,
-            choices=('FILE', 'ZIP'), default='FILE',
+            choices=('FILE', 'ZIP'), default='ZIP',
             help="Output file or a directory when multiple inputs are passed",
             )
     parser.add_argument(
             "-q", "--quiet", dest="use_quiet", action='store_true', required=False,
             help="Suppress status output",
             )
+    parser.add_argument(
+            "-t", "--temp", dest="temp_path", metavar='DIR', required=False,
+            help="Temporary directory to use",
+            )
 
     return parser
 
@@ -563,14 +628,20 @@ def main():
     args = parser.parse_args(sys.argv[1:])
 
     if args.use_quiet:
-        report = lambda msg: None
+        def report(msg):
+            pass
     else:
-        report = lambda msg: print(msg, end="")
+        def report(msg):
+            sys.stdout.write(msg)
+            sys.stdout.flush()
 
     for msg in pack(
             args.path_src.encode('utf-8'),
             args.path_dst.encode('utf-8'),
             mode=args.mode,
+            base_dir_dst_temp=(
+                args.temp_path.encode('utf-8')
+                if args.temp_path else None),
             ):
         report(msg)
 
