@@ -37,10 +37,6 @@ del path
 import logging
 log = logging.getLogger("bam_cli")
 
-# if we're a module, don't mess with logging level
-if __name__ == "__main__":
-    logging.basicConfig(level=logging.DEBUG)
-
 
 def fatal(msg):
     if __name__ == "__main__":
@@ -315,10 +311,11 @@ class bam_session:
             blendfile_abs,  # str
             blendfile,  # bytes
             binary_edits,
+            session_rootdir,
             paths_uuid_update=None,
             ):
 
-        sys.stdout.write("  operating on: %r\n" % blendfile)
+        sys.stdout.write("  operating on: %r\n" % blendfile_abs)
         sys.stdout.flush()
         # we don't want to read, just edit whats there.
         with open(blendfile_abs, 'rb+') as fh_blend:
@@ -337,7 +334,8 @@ class bam_session:
             # update hash!
             # we could do later, but the file is fresh in cache, so do now
             from bam.utils.system import uuid_from_file
-            paths_uuid_update[blendfile.decode('utf-8')] = uuid_from_file(blendfile_abs)
+            f_rel = os.path.relpath(blendfile_abs, session_rootdir)
+            paths_uuid_update[f_rel] = uuid_from_file(blendfile_abs)
             del uuid_from_file
 
     @staticmethod
@@ -355,6 +353,12 @@ class bam_session:
                 assert(not os.path.isabs(path))
                 assert(os.path.exists(os.path.join(session_rootdir, path.decode('utf-8'))))
 
+        with open(os.path.join(session_rootdir, ".bam_paths_remap.json")) as fp:
+            paths_remap = json.load(fp)
+            paths_remap_relbase = paths_remap.get(".", "")
+            paths_remap_reverse = {v: k for k, v in paths_remap.items()}
+            del paths_remap
+
         with open(os.path.join(session_rootdir, ".bam_paths_edit.data"), 'rb') as fh:
             import pickle
             binary_edits_all = pickle.load(fh)
@@ -364,11 +368,16 @@ class bam_session:
                     if paths is not None and blendfile not in paths:
                         continue
 
-                    blendfile_abs = os.path.join(session_rootdir, blendfile.decode('utf-8'))
+                    # get the absolute path as it is in the main repo
+                    # then remap back to our local checkout
+                    blendfile_abs_remote = os.path.normpath(os.path.join(paths_remap_relbase, blendfile.decode('utf-8')))
+                    blendfile_abs = os.path.join(session_rootdir, paths_remap_reverse[blendfile_abs_remote])
+
                     bam_session.binary_edits_apply_single(
                             blendfile_abs,
                             blendfile,
                             binary_edits,
+                            session_rootdir,
                             paths_uuid_update,
                             )
             del pickle
@@ -378,7 +387,6 @@ class bam_session:
             # freshen the UUID's based on the replayed binary_edits
             from bam.utils.system import write_json_to_file
             paths_uuid = bam_session.load_paths_uuid(session_rootdir)
-            assert(set(paths_uuid_update.keys()).issubset(set(paths_uuid.keys())))
             paths_uuid.update(paths_uuid_update)
             write_json_to_file(os.path.join(session_rootdir, ".bam_paths_uuid.json"), paths_uuid)
             del write_json_to_file
@@ -600,7 +608,6 @@ class bam_commands:
             paths_remap = json.load(fp)
 
             paths_uuid = bam_session.load_paths_uuid(session_rootdir)
-            print(paths_uuid)
 
             for f_src, f_dst in paths_remap.items():
                 if f_src == ".":
@@ -1098,13 +1105,13 @@ class bam_commands:
             "file": open(temp_zip, 'rb'),
             }
 
-        r = requests.put(
-                bam_session.request_url("file"),
-                params=payload,
-                auth=(cfg["user"], cfg["password"]),
-                files=files)
+        with files["file"]:
+            r = requests.put(
+                    bam_session.request_url("file"),
+                    params=payload,
+                    auth=(cfg["user"], cfg["password"]),
+                    files=files)
 
-        files["file"].close()
         os.remove(temp_zip)
 
         try:
@@ -1318,9 +1325,12 @@ class bam_commands:
             paths,
             output,
             mode,
+            repository_base_path=None,
             all_deps=False,
             use_quiet=False,
+            warn_remap_externals=False,
             compress_level=-1,
+            filename_filter=None,
             ):
         # Local packing (don't use any project/session stuff)
         from .blend import blendfile_pack
@@ -1343,13 +1353,97 @@ class bam_commands:
         else:
             report = lambda msg: print(msg, end="")
 
+        if repository_base_path is not None:
+            repository_base_path = repository_base_path.encode('utf-8')
+
+        # replace var with a pattern matching callback
+        if filename_filter:
+            # convert string into regex callback
+            # "*.txt;*.png;*.rst" --> r".*\.txt$|.*\.png$|.*\.rst$"
+            import re
+            import fnmatch
+
+            compiled_pattern = re.compile(
+                    b'|'.join(fnmatch.translate(f).encode('utf-8')
+                              for f in filename_filter.split(";") if f),
+                    re.IGNORECASE,
+                    )
+
+            def filename_filter(f):
+                return (not filename_filter.compiled_pattern.match(f))
+            filename_filter.compiled_pattern = compiled_pattern
+
+            del compiled_pattern
+            del re, fnmatch
+
         for msg in blendfile_pack.pack(
                 path.encode('utf-8'),
                 output.encode('utf-8'),
                 mode=mode,
                 all_deps=all_deps,
+                repository_base_path=repository_base_path,
                 compress_level=compress_level,
                 report=report,
+                warn_remap_externals=warn_remap_externals,
+                use_variations=True,
+                filename_filter=filename_filter,
+                ):
+            pass
+
+    @staticmethod
+    def copy(
+            paths,
+            output,
+            base,
+            all_deps=False,
+            use_quiet=False,
+            filename_filter=None,
+            ):
+        # Local packing (don't use any project/session stuff)
+        from .blend import blendfile_copy
+        from bam.utils.system import is_subdir
+
+        paths = [os.path.abspath(path) for path in paths]
+        base = os.path.abspath(base)
+        output = os.path.abspath(output)
+
+        # check all blends are in the base path
+        for path in paths:
+            if not is_subdir(path, base):
+                fatal("Input blend file %r is not a sub directory of %r" % (path, base))
+
+        if use_quiet:
+            report = lambda msg: None
+        else:
+            report = lambda msg: print(msg, end="")
+
+        # replace var with a pattern matching callback
+        if filename_filter:
+            # convert string into regex callback
+            # "*.txt;*.png;*.rst" --> r".*\.txt$|.*\.png$|.*\.rst$"
+            import re
+            import fnmatch
+
+            compiled_pattern = re.compile(
+                    b'|'.join(fnmatch.translate(f).encode('utf-8')
+                              for f in filename_filter.split(";") if f),
+                    re.IGNORECASE,
+                    )
+
+            def filename_filter(f):
+                return (not filename_filter.compiled_pattern.match(f))
+            filename_filter.compiled_pattern = compiled_pattern
+
+            del compiled_pattern
+            del re, fnmatch
+
+        for msg in blendfile_copy.copy_paths(
+                [path.encode('utf-8') for path in paths],
+                output.encode('utf-8'),
+                base.encode('utf-8'),
+                all_deps=all_deps,
+                report=report,
+                filename_filter=filename_filter,
                 ):
             pass
 
@@ -1433,6 +1527,7 @@ def init_argparse_common(
         use_all_deps=False,
         use_quiet=False,
         use_compress_level=False,
+        use_exclude=False,
         ):
     import argparse
 
@@ -1462,6 +1557,20 @@ def init_argparse_common(
                 choices=('default', 'fast', 'best', 'store'),
                 help="Compression level for resulting archive",
                 )
+    if use_exclude:
+        subparse.add_argument(
+                "-e", "--exclude", dest="exclude", metavar='PATTERN(S)', required=False,
+                default="",
+                help="""
+                Optionally exclude files from the pack.
+
+                Using Unix shell-style wildcards *(case insensitive)*.
+                ``--exclude="*.png"``
+
+                Multiple patterns can be passed using the  ``;`` separator.
+                ``--exclude="*.txt;*.avi;*.wav"``
+                """
+                )
 
 
 def create_argparse_init(subparsers):
@@ -1657,6 +1766,15 @@ def create_argparse_pack(subparsers):
 
        # pack a blend with maximum compression for online downloads
        bam pack /path/to/scene.blend --output my_scene.zip --compress=best
+
+    You may also pack a .blend while keeping your whole repository hierarchy by passing
+    the path to the top directory of the repository, and ask to be warned about dependencies paths
+    outside of that base path:
+
+    .. code-block:: sh
+
+       bam pack --repo="/path/to/repo" --warn-external /path/to/repo/path/to/scene.blend
+
     """,
             formatter_class=argparse.RawDescriptionHelpFormatter,
             )
@@ -1674,8 +1792,17 @@ def create_argparse_pack(subparsers):
             choices=('ZIP', 'FILE'),
             help="Output file or a directory when multiple inputs are passed",
             )
+    subparse.add_argument(
+            "--repo", dest="repository_base_path", metavar='DIR', required=False,
+            help="Base directory from which you want to keep existing hierarchy (usually to repository directory),"
+                 "will default to packed blend file's directory if not specified",
+            )
+    subparse.add_argument(
+            "--warn-external", dest="warn_remap_externals", action='store_true',
+            help="Warn for every dependency outside of given repository base path",
+            )
 
-    init_argparse_common(subparse, use_all_deps=True, use_quiet=True, use_compress_level=True)
+    init_argparse_common(subparse, use_all_deps=True, use_quiet=True, use_compress_level=True, use_exclude=True)
 
     subparse.set_defaults(
             func=lambda args:
@@ -1685,10 +1812,62 @@ def create_argparse_pack(subparsers):
                     ((os.path.splitext(args.paths[0])[0] + ".zip")
                      if args.mode == 'ZIP' else None),
                     args.mode,
+                    repository_base_path=args.repository_base_path or None,
                     all_deps=args.all_deps,
                     use_quiet=args.use_quiet,
-                    compress_level=args.compress_level),
-                    )
+                    warn_remap_externals=args.warn_remap_externals,
+                    compress_level=args.compress_level,
+                    filename_filter=args.exclude,
+                    ),
+            )
+
+
+def create_argparse_copy(subparsers):
+    import argparse
+    subparse = subparsers.add_parser(
+            "copy", aliases=("cp",),
+            help="Copy blend file(s) and their dependencies to a new location (maintaining the directory structure).",
+            description=
+    """
+    The line below will copy ``scene.blend`` to ``/destination/to/scene.blend``.
+
+    .. code-block:: sh
+
+       bam copy /path/to/scene.blend --base=/path --output=/destination
+
+    .. code-block:: sh
+
+       # you can also copy multiple files
+       bam copy /path/to/scene.blend /path/other/file.blend --base=/path --output /other/destination
+    """,
+            formatter_class=argparse.RawDescriptionHelpFormatter,
+            )
+    subparse.add_argument(
+            dest="paths", nargs="+",
+            help="Path(s) to blend files to operate on",
+            )
+    subparse.add_argument(
+            "-o", "--output", dest="output", metavar='DIR', required=True,
+            help="Output directory where where files will be copied to",
+            )
+    subparse.add_argument(
+            "-b", "--base", dest="base", metavar='DIR', required=True,
+            help="Base directory for input paths (files outside this path will be omitted)",
+            )
+
+    init_argparse_common(subparse, use_all_deps=True, use_quiet=True, use_exclude=True)
+
+    subparse.set_defaults(
+            func=lambda args:
+            bam_commands.copy(
+                    args.paths,
+                    args.output,
+                    args.base,
+                    all_deps=args.all_deps,
+                    use_quiet=args.use_quiet,
+                    filename_filter=args.exclude,
+                    ),
+            )
 
 
 def create_argparse_remap(subparsers):
@@ -1826,6 +2005,7 @@ def create_argparse():
     # non-bam project commands
     create_argparse_deps(subparsers)
     create_argparse_pack(subparsers)
+    create_argparse_copy(subparsers)
     create_argparse_remap(subparsers)
 
     return parser
@@ -1835,6 +2015,11 @@ def main(argv=None):
     if argv is None:
         argv = sys.argv[1:]
 
+    logging.basicConfig(
+        level=logging.INFO,
+        format='%(asctime)-15s %(levelname)8s %(name)s %(message)s',
+    )
+
     parser = create_argparse()
     args = parser.parse_args(argv)
 
